{--------------------------------------------------------------------}
{ TSMSort V 1.0                                                      }
{ Written by Martin Waldenburg 1996.                                 }
{ Copyright by Martin Waldenburg 1996. All rights reserved.          }
{ This is FreeWare.                                                  }
{ Not testet with Delphi1.                                           }
{ It's provided as is, without a warranty of any kind.               }
{ You use it at your own risc.                                       }
{ You may use and distribute it freely.                              }
{ But You may not say it's your work.                                }
{ If you distribute it you must provide all Files.                   }
{--------------------------------------------------------------------}
{ This demonstrates the implementation and use of a sortable single  }
{ linked list using  MergeSort.                                      }
{ Enjoy it's free.                                                   }
{--------------------------------------------------------------------}

unit mwSMSort;

interface

uses
  SysUtils, Classes;

type
  PLine = ^Line;
  Line = record
            next: PLine;
            key   : string;
          end;

type
  TmwSMSort = class(TObject)
  private
    Line,
    head,
    current: PLine;
    FSize : Integer;
    FSorted,
    FIgnore: boolean;
    Function CompareLines(la, lb : PLine) : Integer;
    Function Merge(a0, lastA, PrevTodoNext : PLine): PLine;
    Function MergeSort(lines : PLine; Count : Integer) : PLine;
    Function CreateLine(n:PLine; item:String):PLine;
  protected
  public
    constructor Create;
    destructor Destroy;override;
    Property Count : integer read FSize write FSize ;
    Procedure Delete(aLine: PLine);
    Procedure DeleteEmptyLines;
    Function Getkey : string;
    Procedure Add(item: String);
    Procedure MoveFirst;
    Procedure MoveNext;
    Procedure LoadFromFile(FileName: String);
    Procedure SaveToFile(FileName: String);
    Procedure Sort;
  published
    Property Sorted : boolean read FSorted write FSorted ;
    Property IgnoreCase : boolean read FIgnore write FIgnore ;
end;

implementation


{ Create a Line with given values}
Function TmwSMSort.CreateLine(n:PLine; item:String):PLine;
var
  newLine: PLine;
begin
  new(Line);
  newLine:= Line;
  with newLine^ do begin
       next:= nil;
       key:= item;
  end;
    Result:= newLine;
end;


{ Create the list with default values }
constructor TmwSMSort.Create;
begin
  inherited Create;
  head:= CreateLine(nil, 'This is the head line');
  current:= head;
  FSize:= 0;
  FSorted:= False;
  FIgnore:= True;
end;


{ Destroy the whole list, Line by Line }
destructor TmwSMSort.Destroy;
begin
   Line := head;
   while Line <> nil do begin
         head:= Line;
         Line := head^.Next;
         Dispose(head);
   end;
   Dispose(Line);
  inherited Destroy;
end;


{ Compare the key of two Lines }
Function TmwSMSort.CompareLines(la, lb : PLine) : Integer;
begin
If FIgnore = False then
  Result:= CompareStr(la^.key, lb^.key)
else
  Result:= CompareText(la^.key, lb^.key);
end;

{ Delete a given Line }
Procedure TmwSMSort.Delete(aLine: PLine);
begin
   if (not(Fsize = 0)) then
     begin
       current^.next:= aLine^.next;
       Dispose(aLine);
       dec(Fsize);
     end;
end;

{ Delete empty Lines }
Procedure TmwSMSort.DeleteEmptyLines;
begin
  MoveFirst;
  while (not(current = nil)) do
    begin
      if current^.key = '' then
        begin
          Delete(current);
        end else
        MoveNext;
    end;
end;

{ Retrive key of current Line. }
function TmwSMSort.Getkey: String;
begin
    Result := current^.key;
end;

{ insert Line after current item, make newLine current }
procedure TmwSMSort.Add(item: String);
begin
    Line:= CreateLine(nil, item);
    current^.next:= Line;
    current:= Line;
    inc(Fsize);
end;


{ make first value in list current }
procedure TmwSMSort.MoveFirst;
begin
    current := head^.next;
end;

{ make next value in list current }
procedure TmwSMSort.MoveNext;
begin
    current := current^.next;
end;

Function TmwSMSort.Merge(a0, lastA, PrevTodoNext : PLine): PLine;
var
  a, b, c: Pline;
begin
  a:= a0^.next;
  b:=lastA^.next;
  while (not(a = b)) and (not(b = PrevTodoNext^.next)) and Assigned(b) do
  begin
    if CompareLines(a, b) <= 0 then
    begin
      a0:= a0^.next;
      a:= a0^.next;
    end else
    begin
      c:= b;
      b:= b^.next;
      a0^.next:= c;
      a0:= a0^.next;
      c^.next:= a;
      lastA^.next:= b;
      if PrevTodoNext = c then begin PrevTodoNext:= lastA; end;
    end;

  end;
  Result:= PrevTodoNext;
end;


Function TmwSMSort.MergeSort(lines : PLine; Count : Integer) : PLine;
var
  PrevFirstLine, a0, lastA, run : PLine;
  N, Na, Nb : Integer;
begin
  N := 1;
  while N < Count do begin
    lastA := nil;
    PrevFirstLine:= Lines;
    run := PrevFirstLine;
    while Assigned(run) do begin
      run := PrevFirstLine;
      a0 := run;
      run := run^.next;
      Na := 0;
      while (Na < N) and Assigned(run) do begin
        Inc(Na);
        lastA := run;
        run := run^.next;
      end;
      Nb := 0;
      while (Nb < N) and Assigned(run) do begin
        Inc(Nb);
        PrevFirstLine := run;
        run := run^.next;
      end;
        PrevFirstLine := merge(a0, lastA, PrevFirstLine);
    end;
     N := N+N;
  end;
    Result := PrevFirstLine;
end;

{ Start sorting }
Procedure TmwSMSort.Sort;
begin
  MergeSort(head, Count);
end;

Procedure TmwSMSort.LoadFromFile(FileName: String);
var
  FromFile: TextFile;
  TextBuffer: String;
begin
  AssignFile(FromFile, FileName);
  reset(FromFile);
  current:= head;
  while not Eof(FromFile) do
    begin
      readln(FromFile, TextBuffer);
      Add(TextBuffer);
    end;
  CloseFile(FromFile);
end;

Procedure TmwSMSort.SaveToFile(FileName: String);
var
  ToFile: TextFile;
  TextBuffer: String;
begin
  AssignFile(ToFile, FileName);
  rewrite(ToFile);
  MoveFirst;
  while not (current = nil) do
    begin
      TextBuffer:= GetKey;
      writeln(ToFile, TextBuffer);
      MoveNext;
    end;
  CloseFile(ToFile);
end;


end.

{--------------------------------------------------------------------}
{ Martin Waldenburg
  Landaeckerstrasse 27
  71642 Ludwigsburg
  Germany
  Share your Code}
{--------------------------------------------------------------------}

